@@ -1,49 +1,96 @@
 #include "gradm.h"
 
+struct file_acl *get_exact_matching_object(struct proc_acl *subject, const char *filename)
+{
+	struct file_acl *tmpf = NULL;
+	struct proc_acl *tmpp = subject;
+	struct file_acl *tmpg = NULL;
+	char *tmpname = alloca(strlen(filename) + 1);
+	int reduced_dir = 0;
+	strcpy(tmpname, filename);
+
+	do {
+		tmpp = subject;
+		do {
+			tmpf = lookup_acl_object_by_name(tmpp, tmpname);
+			if (!tmpf)
+				tmpf = lookup_acl_object_by_inodev(tmpp, tmpname);
+			if (tmpf) {
+				/* check globbed objects */
+				for_each_globbed(tmpg, tmpf) {
+					if (!fnmatch(tmpg->filename, filename, 0))
+						return tmpg;
+				}
+				if (!reduced_dir)
+					return tmpf;
+			}
+		} while ((tmpp = tmpp->parent_subject));
+		reduced_dir = 1;
+	} while (parent_dir(filename, &tmpname));
+
+	// won't get here
+	return NULL;
+}
+
+static struct file_acl *get_a_matching_object(struct proc_acl *subject, 
+				const char *filename, const char *origname)
+{
+	struct file_acl *tmpf, *tmpg;
+	struct proc_acl *tmpp = subject;
+	do {
+		tmpf = lookup_acl_object_by_name(tmpp, filename);
+		if (!tmpf)
+			tmpf = lookup_acl_object_by_inodev(tmpp, filename);
+		if (tmpf) {
+			/* check globbed objects */
+			for_each_globbed(tmpg, tmpf) {
+				if (!fnmatch(tmpg->filename, origname, 0))
+					return tmpg;
+			}
+			return tmpf;
+		}
+	} while ((tmpp = tmpp->parent_subject));
+
+	return NULL;
+}
+
+struct file_acl *get_matching_object(struct proc_acl *subject, const char *filename)
+{
+	struct file_acl *tmpf = NULL;
+	struct file_acl *tmpg = NULL;
+	char *tmpname = alloca(strlen(filename) + 1);
+
+	strcpy(tmpname, filename);
+
+	do {
+		tmpf = get_a_matching_object(subject, tmpname, filename);
+		if (tmpf)
+			return tmpf;
+	} while (parent_dir(filename, &tmpname));
+
+	// won't get here
+	return NULL;
+}
+
 static int
 check_permission(struct role_acl *role, struct proc_acl *def_acl,
 		 const char *filename, struct chk_perm *chk)
 {
 	struct file_acl *tmpf = NULL;
 	struct proc_acl *tmpp = def_acl;
-	struct file_acl *tmpg = NULL;
-	char *tmpname;
 	gr_cap_t cap_drp = {{ 0, 0 }}, cap_mask = {{ 0, 0 }};
 	gr_cap_t cap_full = {{ ~0, ~0 }};
 
 	if (chk->type == CHK_FILE) {
-		tmpname = alloca(strlen(filename) + 1);
-		strcpy(tmpname, filename);
-
-		do {
-			tmpp = def_acl;
-			do {
-				tmpf = lookup_acl_object_by_name(tmpp, tmpname);
-				if (tmpf) {
-					/* check globbed objects */
-					for_each_globbed(tmpg, tmpf) {
-						if (!fnmatch(tmpg->filename, filename, 0)) {
-							if (((chk->w_modes == 0xffff)
-							     || (tmpg->mode & chk->w_modes))
-							    && ((chk->u_modes == 0xffff)
-								|| !(tmpg->mode & chk->u_modes))) {
-								return 1;
-							} else {
-								return 0;
-							}
-						}
-					}
-					if (((chk->w_modes == 0xffff)
-					     || (tmpf->mode & chk->w_modes))
-					    && ((chk->u_modes == 0xffff)
-						|| !(tmpf->mode & chk->u_modes))) {
-						return 1;
-					} else {
-						return 0;
-					}
-				}
-			} while ((tmpp = tmpp->parent_subject));
-		} while (parent_dir(filename, &tmpname));
+		tmpf = get_matching_object(def_acl, filename);
+		if (((chk->w_modes == 0xffff)
+		     || (tmpf->mode & chk->w_modes))
+		     && ((chk->u_modes == 0xffff)
+		     || !(tmpf->mode & chk->u_modes))) {
+			return 1;
+		} else {
+			return 0;
+		}
 	} else if (chk->type == CHK_CAP) {
 		cap_mask = tmpp->cap_mask;
 		cap_drp = tmpp->cap_drop;
@@ -64,6 +111,45 @@ check_permission(struct role_acl *role, struct proc_acl *def_acl,
 	return 0;
 }
 
+static void
+insert_globbed_objects(void)
+{
+	struct glob_file *glob;
+
+	for (glob = glob_files_head; glob; glob = glob->next) {
+		add_globbed_object_acl(glob->subj, glob->filename, glob->mode, glob->type, glob->policy_file, glob->lineno);
+	}
+
+	return;
+}
+
+static void
+check_symlinks(void)
+{
+	struct symlink *sym;
+	struct file_acl *tmpf;
+
+	for (sym = symlinks; sym; sym = sym->next) {
+		char buf[PATH_MAX];
+		memset(&buf, 0, sizeof (buf));
+
+		if (!realpath(sym->obj->filename, buf))
+			continue;
+
+		/* warning exemptions */
+		if (!strcmp(buf, "/proc/self"))
+			continue;
+
+		tmpf = get_exact_matching_object(sym->subj, buf);
+		if (tmpf == NULL) {
+			fprintf(stdout, "Warning: object does not exist in role %s, subject %s for the target of the symlink object %s specified on line %lu of %s.\n",
+				sym->role->rolename, sym->subj->filename, sym->obj->filename, sym->lineno, sym->policy_file);
+		}
+	}
+
+	return;
+}
+
 static int
 check_subjects(struct role_acl *role)
 {
@@ -335,22 +421,19 @@ handle_notrojan_mode(void)
 			for_each_object(obj, subj) {
 				if (!(obj->mode & GR_EXEC))
 					continue;
-				if ((objname =
-				     malloc(strlen(obj->filename) + 1)) == NULL)
-					failure("malloc");
-				strcpy(objname, obj->filename);
-				do {
-					for_each_role(role2, current_role) {
-						if (!strcmp(role2->rolename, ":::kernel:::"))
+				for_each_role(role2, current_role) {
+					if (!strcmp(role2->rolename, ":::kernel:::"))
+						continue;
+					if (role2->roletype & GR_ROLE_GOD)
+						continue;
+					for_each_subject(subj2, role2) {
+						if (subj2 == subj
+						    || (subj2->
+							filename[0] !=
+							'/'))
 							continue;
-						if (role2->roletype & GR_ROLE_GOD)
-							continue;
-						for_each_subject(subj2, role2) {
-							if (subj2 == subj
-							    || (subj2->
-								filename[0] !=
-								'/'))
-								continue;
+						objname = gr_strdup(obj->filename);
+						do {
 							obj2 = lookup_acl_object_by_name(subj2, objname);
 							if (obj2 && obj2->mode & GR_WRITE) {
 								ret++;
@@ -365,11 +448,17 @@ handle_notrojan_mode(void)
 								     role2->rolename, subj2->filename,
 								     obj2->filename, subj->filename);
 								break;
+							} else if (obj2) {
+								/* if we found a match, but it wasn't writable, then just break
+								   otherwise we'd end up matching on a directory whose
+								   mode doesn't apply to what we're looking up
+								*/
+								break;
 							}
-						}
+						} while (parent_dir(obj->filename, &objname));
+						free(objname);
 					}
-				} while (parent_dir(obj->filename, &objname));
-				free(objname);
+				}
 			}
 		}
 	}
@@ -463,6 +552,8 @@ analyze_acls(void)
 	struct stat fstat;
 	gr_cap_t cap_full = {{ ~0, ~0 }};
 
+	insert_globbed_objects();
+
 	errs_found = check_role_transitions();
 
 	for_each_role(role, current_role)
@@ -478,6 +569,21 @@ analyze_acls(void)
 	}
 
 	for_each_role(role, current_role) {
+		if (((role->roletype & (GR_ROLE_GOD | GR_ROLE_PERSIST)) == 
+		     (GR_ROLE_GOD | GR_ROLE_PERSIST)) &&
+		    !strcmp(role->rolename, "admin")) {
+			fprintf(stderr, "The admin role has been marked "
+			"as a persistent role.  This severely compromises "
+			"security as any process restarted via an admin "
+			"role will retain the admin role indefinitely.\n"
+			"Please create a specific role for the handling "
+			"of system shutdown (the common use case of "
+			"persistent special roles).  The RBAC system will "
+			"not be allowed to be enabled until this error is "
+			"fixed.\n");
+			exit(EXIT_FAILURE);
+		}
+
 		def_acl = role->root_label;
 		if (!def_acl) {
 			fprintf(stderr, "There is no default subject for "
@@ -561,6 +667,14 @@ analyze_acls(void)
 			errs_found++;
 		}
 
+		if (!check_permission(role, def_acl, "/run", &chk)) {
+			fprintf(stderr,
+				"Writing access is allowed by role %s to /run, the directory which "
+				"holds information for running services and potentially the initctl device.\n\n",
+				role->rolename);
+			errs_found++;
+		}
+
 		if (!stat("/lib/modules", &fstat) && !check_permission(role, def_acl, "/lib/modules", &chk)) {
 			fprintf(stderr,
 				"Writing access is allowed by role %s to /lib/modules, the directory which "
@@ -569,6 +683,14 @@ analyze_acls(void)
 			errs_found++;
 		}
 
+		if (!stat("/lib64/modules", &fstat) && !check_permission(role, def_acl, "/lib64/modules", &chk)) {
+			fprintf(stderr,
+				"Writing access is allowed by role %s to /lib64/modules, the directory which "
+				"holds kernel modules.\n\n",
+				role->rolename);
+			errs_found++;
+		}
+
 		if (!check_permission(role, def_acl, "/dev", &chk)) {
 			fprintf(stderr,
 				"Writing access is allowed by role %s to /dev, the directory which "
@@ -640,6 +762,22 @@ analyze_acls(void)
 			errs_found++;
 		}
 
+		if (!stat("/lib32", &fstat) && !check_permission(role, def_acl, "/lib32", &chk)) {
+			fprintf(stderr,
+				"Write access is allowed by role %s to /lib32, a directory which "
+				"holds system libraries.\n\n",
+				role->rolename);
+			errs_found++;
+		}
+
+		if (!stat("/usr/lib32", &fstat) && !check_permission(role, def_acl, "/usr/lib32", &chk)) {
+			fprintf(stderr,
+				"Write access is allowed by role %s to /usr/lib32, a directory which "
+				"holds system libraries.\n\n", role->rolename);
+			errs_found++;
+		}
+
+
 		if (!stat("/lib64", &fstat) && !check_permission(role, def_acl, "/lib64", &chk)) {
 			fprintf(stderr,
 				"Write access is allowed by role %s to /lib64, a directory which "
@@ -664,6 +802,14 @@ analyze_acls(void)
 			errs_found++;
 		}
 
+		if (!stat("/sys", &fstat) && !check_permission(role, def_acl, "/sys", &chk)) {
+			fprintf(stderr,
+				"Read access is allowed by role %s to /sys, the directory which "
+				"holds entries that often leak information from the kernel.\n\n",
+				role->rolename);
+			errs_found++;
+		}
+
 		if (!stat("/proc/slabinfo", &fstat) && !check_permission(role, def_acl, "/proc/slabinfo", &chk)) {
 			fprintf(stderr,
 				"Reading access is allowed by role %s to "
@@ -708,6 +854,30 @@ analyze_acls(void)
 			errs_found++;
 		}
 
+		if (!stat("/lib32/modules", &fstat) && !check_permission(role, def_acl, "/lib32/modules", &chk)) {
+			fprintf(stderr,
+				"Reading access is allowed by role %s to "
+				"/lib32/modules, the directory which holds kernel "
+				"kernel modules.  The ability to read these "
+				"images provides an attacker with very "
+				"useful information for launching \"ret-to-libc\" "
+				"style attacks against the kernel"
+				".\n\n", role->rolename);
+			errs_found++;
+		}
+
+		if (!stat("/lib64/modules", &fstat) && !check_permission(role, def_acl, "/lib64/modules", &chk)) {
+			fprintf(stderr,
+				"Reading access is allowed by role %s to "
+				"/lib64/modules, the directory which holds kernel "
+				"kernel modules.  The ability to read these "
+				"images provides an attacker with very "
+				"useful information for launching \"ret-to-libc\" "
+				"style attacks against the kernel"
+				".\n\n", role->rolename);
+			errs_found++;
+		}
+
 		if (!stat("/proc/kallsyms", &fstat) && !check_permission(role, def_acl, "/proc/kallsyms", &chk)) {
 			fprintf(stderr,
 				"Reading access is allowed by role %s to "
@@ -746,6 +916,16 @@ analyze_acls(void)
 			errs_found++;
 		}
 
+		chk.u_caps = cap_conv("CAP_SYSLOG");
+		chk.w_caps = cap_full;
+
+		if (!check_permission(role, def_acl, "", &chk)) {
+			fprintf(stderr, "CAP_SYSLOG is not "
+				"removed in role %s.  This would allow an "
+				"attacker to view OOPs messages in dmesg that contain addresses useful for kernel exploitation.\n\n",
+				role->rolename);
+			errs_found++;
+		}
 
 		chk.u_caps = cap_conv("CAP_SYS_BOOT");
 		chk.w_caps = cap_full;
@@ -794,6 +974,18 @@ analyze_acls(void)
 		}
 
 
+		chk.u_caps = cap_conv("CAP_SETFCAP");
+		chk.w_caps = cap_full;
+
+		if (!check_permission(role, def_acl, "", &chk)) {
+			fprintf(stderr, "CAP_SETFCAP is not "
+				"removed for role %s.  This would allow an "
+				"attacker to set and modify file "
+				"capabilities.\n\n", role->rolename);
+			errs_found++;
+		}
+
+
 		errs_found += check_path_env(role, def_acl);
 		errs_found += check_lib_paths(role, def_acl);
 		errs_found += check_lilo_conf(role, def_acl);
@@ -802,6 +994,8 @@ analyze_acls(void)
 
 	errs_found += handle_notrojan_mode();
 
+	check_symlinks();
+
 	if (errs_found) {
 		printf("There were %d holes found in your RBAC "
 		       "configuration.  These must be fixed before the "
