@@ -11,9 +11,11 @@
 #define GR_POLICY_PATH 		GRSEC_DIR "/policy"
 #define GR_PW_PATH 		GRSEC_DIR "/pw"
 #define GR_LEARN_CONFIG_PATH	GRSEC_DIR "/learn_config"
+#define GR_LEARN_PIPE_PATH	GRSEC_DIR "/.grlearn.pipe"
+#define GR_LEARN_PID_PATH	GRSEC_DIR "/.grlearn.pid"
 
-#define GR_VERSION		"2.1.14"
-#define GRADM_VERSION		0x2114
+#define GR_VERSION		"2.9"
+#define GRADM_VERSION		0x2900
 
 #define GR_PWONLY		0
 #define GR_PWANDSUM		1
@@ -27,7 +29,6 @@
 #define GR_FEXIST		0x1
 #define GR_FFAKE		0x2
 #define GR_FLEARN		0x4
-#define GR_SYMLINK		0x8
 
 #define CHK_FILE		0
 #define CHK_CAP			1
@@ -36,6 +37,9 @@
 #define PATH_MAX 		4096
 #define MAX_LINE_LEN 		5000
 
+// CAP_WAKE_ALARM
+#define CAP_MAX			35
+
 #define MAX_INCLUDE_DEPTH	20
 #define MAX_NEST_DEPTH		8
 #define MAX_SYMLINK_DEPTH	8
@@ -59,6 +63,54 @@
 #define RLIMIT_RTTIME 15
 #endif
 
+#ifndef AF_RDS
+#define AF_RDS 21
+#endif
+#ifndef AF_SNA
+#define AF_SNA 22
+#endif
+#ifndef AF_IRDA
+#define AF_IRDA 23
+#endif
+#ifndef AF_PPOX
+#define AF_PPOX 24
+#endif
+#ifndef AF_WANPIPE
+#define AF_WANPIPE 25
+#endif
+#ifndef AF_LLC
+#define AF_LLC 26
+#endif
+#ifndef AF_CAN
+#define AF_CAN 29
+#endif
+#ifndef AF_TIPC
+#define AF_TIPC 30
+#endif
+#ifndef AF_BLUETOOTH
+#define AF_BLUETOOTH 31
+#endif
+#ifndef AF_IUCV
+#define AF_IUCV 32
+#endif
+#ifndef AF_RXRPC
+#define AF_RXRPC 33
+#endif
+#ifndef AF_ISDN
+#define AF_ISDN 34
+#endif
+#ifndef AF_PHONET
+#define AF_PHONET 35
+#endif
+#ifndef AF_IEEE802154
+#define AF_IEEE802154 36
+#endif
+#ifndef AF_CAIF
+#define AF_CAIF 37
+#endif
+#undef AF_MAX
+#define AF_MAX 38
+
 #define GR_NLIMITS	32
 #define GR_CRASH_RES	31
 
@@ -90,6 +142,7 @@ enum {
 	GR_IP_BIND 	= 0x01,
 	GR_IP_CONNECT 	= 0x02,
 	GR_IP_INVERT 	= 0x04,
+	GR_SOCK_FAMILY  = 0x20
 };
 
 enum {
@@ -125,7 +178,9 @@ enum {
 	GR_AUDIT_LINK	= 0x00080000,
 	GR_PTRACERD 	= 0x00100000,
 	GR_NOPTRACE	= 0x00200000,
-	GR_SUPPRESS 	= 0x00400000
+	GR_SUPPRESS 	= 0x00400000,
+	GR_NOLEARN	= 0x00800000,
+	GR_INIT_TRANSFER= 0x01000000
 };
 
 enum {
@@ -140,6 +195,8 @@ enum {
 	GR_ROLE_TPE 	= 0x0100,
 	GR_ROLE_DOMAIN 	= 0x0200,
 	GR_ROLE_PAM 	= 0x0400,
+	GR_ROLE_PERSIST = 0x0800,
+	GR_ROLE_ISID	= 0x1000
 };
 
 enum {
@@ -164,7 +221,15 @@ enum {
 	GR_INHERITLEARN = 0x00004000,
 	GR_PROCFIND	= 0x00008000,
 	GR_POVERRIDE	= 0x00010000,
-	GR_KERNELAUTH	= 0x00020000
+	GR_KERNELAUTH	= 0x00020000,
+	GR_ATSECURE	= 0x00040000,
+	GR_SHMEXEC	= 0x00080000,
+	GR_GLOBANCHOR	= 0x00100000
+};
+
+enum {
+	GR_DONT_LEARN_ALLOWED_IPS = 0x00000001,
+	GR_SPLIT_ROLES = 0x00000002
 };
 
 /* internal use only.  not to be modified */
@@ -178,6 +243,11 @@ struct capability_set {
 	int cap_val;
 };
 
+struct family_set {
+	char *family_name;
+	int family_val;
+};
+
 struct paxflag_set {
 	char *paxflag_name;
 	u_int16_t paxflag_val;
@@ -264,6 +334,8 @@ struct role_acl {
 	uid_t *domain_children;
 	u_int16_t domain_child_num;
 
+	u_int16_t  umask;
+
 	struct proc_acl **subj_hash;
 	u_int32_t subj_hash_size;
 };
@@ -275,6 +347,7 @@ struct proc_acl {
 	u_int32_t mode;
 	gr_cap_t cap_mask;
 	gr_cap_t cap_drop;
+	gr_cap_t cap_invert_audit;
 
 	struct rlimit res[GR_NLIMITS];
 	u_int32_t resmask;
@@ -286,6 +359,7 @@ struct proc_acl {
 	u_int16_t user_trans_num;
 	u_int16_t group_trans_num;
 
+	u_int32_t sock_families[2];
 	u_int32_t ip_proto[8];
 	u_int32_t ip_type;
 	struct ip_acl **ips;
@@ -306,6 +380,8 @@ struct proc_acl {
 };
 
 struct gr_learn_ip_node {
+	struct gr_learn_ip_node *prev;
+	struct gr_learn_ip_node *next;
 	u_int8_t ip_node;
 	u_int16_t **ports;
 	u_int32_t ip_proto[8];
@@ -314,10 +390,12 @@ struct gr_learn_ip_node {
 	unsigned char all_low_ports:1;
 	unsigned char all_high_ports:1;
 	struct gr_learn_ip_node *parent;
-	struct gr_learn_ip_node **leaves;
+	struct gr_learn_ip_node *leaves;
 };
 
 struct gr_learn_role_entry {
+	struct gr_learn_role_entry *prev;
+	struct gr_learn_role_entry *next;
 	char *rolename;
 	u_int16_t rolemode;
 	unsigned int id;
@@ -327,9 +405,11 @@ struct gr_learn_role_entry {
 };	
 
 struct gr_learn_group_node {
+	struct gr_learn_group_node *prev;
+	struct gr_learn_group_node *next;
 	char *rolename;
 	gid_t gid;
-	struct gr_learn_user_node **users;
+	struct gr_learn_user_node *users;
 	struct gr_hash_struct *hash;
 	struct gr_learn_file_node *subject_list;
 	struct gr_learn_ip_node *allowed_ips;
@@ -342,6 +422,8 @@ struct gr_learn_file_tmp_node {
 };
 
 struct gr_learn_user_node {
+	struct gr_learn_user_node *prev;
+	struct gr_learn_user_node *next;
 	char *rolename;
 	uid_t uid;
 	int multgroups;
@@ -357,12 +439,15 @@ struct gr_learn_subject_node {
 	u_int32_t resmask;
 	u_int16_t pax_flags;
 	u_int32_t inaddr_any_override;
+	u_int32_t sock_families[2];
 };
 
 struct gr_learn_file_node {
+	struct gr_learn_file_node *prev;
+	struct gr_learn_file_node *next;
 	char *filename;
 	u_int32_t mode;
-	struct gr_learn_file_node **leaves;
+	struct gr_learn_file_node *leaves;
 	struct gr_learn_file_node *parent;
 	struct gr_hash_struct *hash;
 	struct gr_learn_file_node *object_list;
@@ -395,6 +480,39 @@ struct deleted_file {
 	struct deleted_file *next;
 };
 
+/* to keep track of symlinks, for processing after all other objects have
+   been added
+*/
+
+struct symlink {
+	struct role_acl *role;
+	struct file_acl *obj;
+	struct proc_acl *subj;
+	char *policy_file;
+	unsigned long lineno;
+	struct symlink *next;
+};
+
+/* to keep track of globbed files, so that the ordering of their anchor
+   doesn't matter
+*/
+
+struct glob_file {
+	struct role_acl *role;
+	struct proc_acl *subj;
+	char *filename;
+	u_int32_t mode;
+	int type;
+	char *policy_file;
+	unsigned long lineno;
+	struct glob_file *next;
+};
+
+extern struct glob_file *glob_files_head;
+extern struct glob_file *glob_files_tail;
+
+extern struct symlink *symlinks;
+
 extern struct deleted_file *deleted_files;
 
 extern unsigned long lineno;
@@ -455,8 +573,9 @@ struct gr_arg_wrapper {
 };
 
 extern char *rlim_table[GR_NLIMITS];
-extern struct capability_set capability_list[35];
+extern struct capability_set capability_list[CAP_MAX+2];
 extern struct paxflag_set paxflag_list[5];
+extern struct family_set sock_families[AF_MAX+2];
 
 extern int is_24_kernel;
 
@@ -475,9 +594,13 @@ extern u_int16_t current_learn_rolemode;
 extern char **dont_reduce_dirs;
 extern char **always_reduce_dirs;
 extern char **protected_paths;
+extern char **read_protected_paths;
 extern char **high_reduce_dirs;
 extern char **high_protected_paths;
+extern u_int32_t grlearn_options;
 
 extern int gr_learn;
+extern int gr_fulllearn;
+extern char *output_log;
 
 #endif
