@@ -3,12 +3,68 @@
 extern FILE *gradmin;
 extern int gradmparse(void);
 
+void set_role_umask(struct role_acl *role, u_int16_t umask)
+{
+	role->umask = umask;
+}
+
+static int get_id_from_role_name(const char *rolename, u_int16_t type)
+{
+	unsigned long the_id = 0;
+	struct passwd *pwd;
+	struct group *grp;
+	char *endptr;
+	
+	if (type & GR_ROLE_USER) {
+		pwd = getpwnam(rolename);
+
+		if (!pwd) {
+			/* now try it as a uid */
+			the_id = strtoul(rolename, &endptr, 10);
+			if (*endptr == '\0')
+				pwd = getpwuid((int)the_id);
+			if (the_id > INT_MAX || *endptr != '\0') {
+				fprintf(stderr, "User %s on line %lu of %s "
+					"is invalid.\nThe RBAC system will "
+					"not be allowed to be enabled until "
+					"this error is fixed.\n", rolename,
+					lineno, current_acl_file);
+				exit(EXIT_FAILURE);
+			}
+		}
+		if (pwd)
+			the_id = pwd->pw_uid;
+		/* else, the_id obtained above via strtoul is valid */
+	} else if (type & GR_ROLE_GROUP) {
+		grp = getgrnam(rolename);
+
+		if (!grp) {
+			/* now try it as a gid */
+			the_id = strtoul(rolename, &endptr, 10);
+			if (*endptr == '\0')
+				grp = getgrgid((int)the_id);
+			if (the_id > INT_MAX || *endptr != '\0') {
+				fprintf(stderr, "Group %s on line %lu of %s "
+					"is invalid.\nThe RBAC system will "
+					"not be allowed to be enabled until "
+					"this error is fixed.\n", rolename,
+					lineno, current_acl_file);
+				exit(EXIT_FAILURE);
+			}
+		}
+		if (grp)
+			the_id = grp->gr_gid;
+		/* else, the_id obtained above via strtoul is valid */
+	}
+
+	return (int)the_id;
+}
+
 void
 add_id_transition(struct proc_acl *subject, char *idname, int usergroup, int allowdeny)
 {
-	struct passwd *pwd;
-	struct group *grp;
 	int i;
+	int id;
 
 	if (usergroup == GR_ID_USER) {
 		if ((subject->user_trans_type | allowdeny) == (GR_ID_ALLOW | GR_ID_DENY)) {
@@ -20,36 +76,20 @@ add_id_transition(struct proc_acl *subject, char *idname, int usergroup, int all
 		}
 		subject->user_trans_type |= allowdeny;
 
+		id = get_id_from_role_name(idname, GR_ROLE_USER);
+
 		/* dupecheck */
 		for (i = 0; i < subject->user_trans_num; i++)
-			if (*(subject->user_transitions + i) == usergroup)
+			if (subject->user_transitions[i] == id)
 				return;
 
-		pwd = getpwnam(idname);
-		if (!pwd) {
-			/* now try it as a uid */
-			unsigned long theuid = 0;
-			char *endptr;
-			theuid = strtoul(idname, &endptr, 10);
-			if (*endptr == '\0')
-				pwd = getpwuid((int)theuid);
-			if (!pwd || theuid > INT_MAX) {
-				fprintf(stderr, "User %s on line %lu of %s "
-					"does not exist.\nThe RBAC system will "
-					"not be allowed to be enabled until "
-					"this error is fixed.\n", idname,
-					lineno, current_acl_file);
-				exit(EXIT_FAILURE);
-			}
-		}
-
 		/* increment pointer count upon allocation of user transition list */
 		if (subject->user_transitions == NULL)
 			num_pointers++;
 
 		subject->user_trans_num++;
 		subject->user_transitions = gr_dyn_realloc(subject->user_transitions, subject->user_trans_num * sizeof(uid_t));
-		*(subject->user_transitions + subject->user_trans_num - 1) = pwd->pw_uid;
+		subject->user_transitions[subject->user_trans_num - 1] = id;
 	} else if (usergroup == GR_ID_GROUP) {
 		if ((subject->group_trans_type | allowdeny) == (GR_ID_ALLOW | GR_ID_DENY)) {
 			fprintf(stderr, "Error on line %lu of %s.  You cannot use "
@@ -60,37 +100,20 @@ add_id_transition(struct proc_acl *subject, char *idname, int usergroup, int all
 		}
 		subject->group_trans_type |= allowdeny;
 
+		id = get_id_from_role_name(idname, GR_ROLE_GROUP);
+
 		/* dupecheck */
 		for (i = 0; i < subject->group_trans_num; i++)
-			if (*(subject->group_transitions + i) == usergroup)
+			if (subject->group_transitions[i] == id)
 				return;
 
-		grp = getgrnam(idname);
-
-		if (!grp) {
-			/* now try it as a gid */
-			unsigned long thegid = 0;
-			char *endptr;
-			thegid = strtoul(idname, &endptr, 10);
-			if (*endptr == '\0')
-				grp = getgrgid((int)thegid);
-			if (!grp || thegid > INT_MAX) {
-				fprintf(stderr, "Group %s on line %lu of %s "
-					"does not exist.\nThe RBAC system will "
-					"not be allowed to be enabled until "
-					"this error is fixed.\n", idname,
-					lineno, current_acl_file);
-				exit(EXIT_FAILURE);
-			}
-		}
-
 		/* increment pointer count upon allocation of group transition list */
 		if (subject->group_transitions == NULL)
 			num_pointers++;
 
 		subject->group_trans_num++;
 		subject->group_transitions = gr_dyn_realloc(subject->group_transitions, subject->group_trans_num * sizeof(gid_t));
-		*(subject->group_transitions + subject->group_trans_num - 1) = grp->gr_gid;
+		subject->group_transitions[subject->group_trans_num - 1] = id;
 	}
 
 	return;
@@ -100,10 +123,23 @@ static int
 is_role_dupe(struct role_acl *role, const char *rolename, const u_int16_t type)
 {
 	struct role_acl *tmp;
+	int id;
+	int i;
 
-	for_each_role(tmp, role)
-	    if ((tmp->roletype & (GR_ROLE_USER | GR_ROLE_GROUP | GR_ROLE_SPECIAL) & type) && !strcmp(tmp->rolename, rolename))
-		return 1;
+	if ((type & GR_ROLE_ISID) || ((type & (GR_ROLE_USER | GR_ROLE_GROUP)) && !(type & GR_ROLE_DOMAIN))) {
+		id = get_id_from_role_name(rolename, type);
+	}
+
+	for_each_role(tmp, role) {
+		if ((tmp->roletype & (GR_ROLE_USER | GR_ROLE_GROUP | GR_ROLE_SPECIAL) & type) && !strcmp(tmp->rolename, rolename))
+			return 1;
+		if ((tmp->roletype & GR_ROLE_DOMAIN) && (type & (GR_ROLE_USER | GR_ROLE_GROUP))) {
+			for (i = 0; i < tmp->domain_child_num; i++) {
+				if (tmp->domain_children[i] == id)
+					return 1;
+			}
+	    	}
+	}
 
 	return 0;
 }
@@ -111,10 +147,13 @@ is_role_dupe(struct role_acl *role, const char *rolename, const u_int16_t type)
 void
 add_domain_child(struct role_acl *role, char *idname)
 {
-	struct passwd *pwd;
-	struct group *grp;
+	if (!(role->roletype & (GR_ROLE_USER | GR_ROLE_GROUP))) {
+		// should never get here
+		fprintf(stderr, "Unhandled exception 1.\n");
+		exit(EXIT_FAILURE);
+	}
 
-	if (is_role_dupe(current_role, idname, role->roletype)) {
+	if (is_role_dupe(current_role, idname, role->roletype | GR_ROLE_ISID)) {
 		fprintf(stderr, "Duplicate role %s on line %lu of %s.\n"
 			"The RBAC system will not be allowed to be "
 			"enabled until this error is fixed.\n",
@@ -133,57 +172,9 @@ add_domain_child(struct role_acl *role, char *idname)
 	if (role->domain_children == NULL)
 		num_pointers++;
 
-	if (role->roletype & GR_ROLE_USER) {
-		pwd = getpwnam(idname);
-
-		if (!pwd) {
-			/* now try it as a uid */
-			unsigned long theuid = 0;
-			char *endptr;
-			theuid = strtoul(idname, &endptr, 10);
-			if (*endptr == '\0')
-				pwd = getpwuid((int)theuid);
-			if (!pwd || theuid > INT_MAX) {
-				fprintf(stderr, "User %s on line %lu of %s "
-					"does not exist.\nThe RBAC system will "
-					"not be allowed to be enabled until "
-					"this error is fixed.\n", idname,
-					lineno, current_acl_file);
-				exit(EXIT_FAILURE);
-			}
-		}
-
-		role->domain_child_num++;
-		role->domain_children = gr_dyn_realloc(role->domain_children, role->domain_child_num * sizeof(uid_t));
-		*(role->domain_children + role->domain_child_num - 1) = pwd->pw_uid;
-	} else if (role->roletype & GR_ROLE_GROUP) {
-		grp = getgrnam(idname);
-
-		if (!grp) {
-			/* now try it as a gid */
-			unsigned long thegid = 0;
-			char *endptr;
-			thegid = strtoul(idname, &endptr, 10);
-			if (*endptr == '\0')
-				grp = getgrgid((int)thegid);
-			if (!grp || thegid > INT_MAX) {
-				fprintf(stderr, "Group %s on line %lu of %s "
-					"does not exist.\nThe RBAC system will "
-					"not be allowed to be enabled until "
-					"this error is fixed.\n", idname,
-					lineno, current_acl_file);
-				exit(EXIT_FAILURE);
-			}
-		}
-
-		role->domain_child_num++;
-		role->domain_children = gr_dyn_realloc(role->domain_children, role->domain_child_num * sizeof(uid_t));
-		*(role->domain_children + role->domain_child_num - 1) = grp->gr_gid;
-	} else {
-		// should never get here
-		fprintf(stderr, "Unhandled exception 1.\n");
-		exit(EXIT_FAILURE);
-	}
+	role->domain_child_num++;
+	role->domain_children = gr_dyn_realloc(role->domain_children, role->domain_child_num * sizeof(gid_t));
+	*(role->domain_children + role->domain_child_num - 1) = get_id_from_role_name(idname, role->roletype);
 
 	return;
 }
@@ -217,17 +208,33 @@ add_role_transition(struct role_acl *role, char *rolename)
 	return;
 }
 
+void add_symlink(struct proc_acl *subj, struct file_acl *obj)
+{
+	struct symlink *sym = malloc(sizeof (struct symlink));
+	if (!sym)
+		failure("malloc");
+
+	sym->role = current_role;
+	sym->subj = subj;
+	sym->obj = obj;
+	sym->policy_file = current_acl_file;
+	sym->lineno = lineno;
+
+	sym->next = symlinks;
+	symlinks = sym;
+
+	return;
+}
+
 static struct deleted_file *
 is_deleted_file_dupe(const char *filename)
 {
 	struct deleted_file *tmp;
 
-	tmp = deleted_files;
-
-	do {
+	for (tmp = deleted_files; tmp; tmp = tmp->next) {
 		if (!strcmp(filename, tmp->filename))
 			return tmp;
-	} while ((tmp = tmp->next));
+	}
 
 	return NULL;
 }
@@ -241,25 +248,16 @@ add_deleted_file(char *filename)
 
 	ino++;
 
-	if (!deleted_files) {
-		deleted_files = malloc(sizeof (struct deleted_file));
-		if (!deleted_files)
-			failure("malloc");
-		deleted_files->filename = filename;
-		deleted_files->ino = ino;
-		deleted_files->next = NULL;
-	} else {
-		retfile = is_deleted_file_dupe(filename);
-		if (retfile)
-			return retfile;
-		dfile = malloc(sizeof (struct deleted_file));
-		if (!dfile)
-			failure("malloc");
-		dfile->filename = filename;
-		dfile->ino = ino;
-		dfile->next = deleted_files;
-		deleted_files = dfile;
-	}
+	retfile = is_deleted_file_dupe(filename);
+	if (retfile)
+		return retfile;
+	dfile = malloc(sizeof (struct deleted_file));
+	if (!dfile)
+		failure("malloc");
+	dfile->filename = filename;
+	dfile->ino = ++ino;
+	dfile->next = deleted_files;
+	deleted_files = dfile;
 
 	return deleted_files;
 }
@@ -298,8 +296,17 @@ int
 add_role_acl(struct role_acl **role, char *rolename, u_int16_t type, int ignore)
 {
 	struct role_acl *rtmp;
-	struct passwd *pwd;
-	struct group *grp;
+
+	if (current_role && current_role->hash == NULL) {
+		fprintf(stderr, "Error on line %lu of %s: "
+				"Attempting to add the role \"%s\" when "
+				"no subjects have been specified for "
+				"the previous role \"%s\".\nThe RBAC "
+				"system will not be allowed to be "
+				"enabled until this error is fixed.\n", 
+			lineno, current_acl_file, rolename, current_role->rolename);
+		exit(EXIT_FAILURE);
+	}
 
 	num_roles++;
 
@@ -315,6 +322,7 @@ add_role_acl(struct role_acl **role, char *rolename, u_int16_t type, int ignore)
 	     (struct role_acl *) calloc(1, sizeof (struct role_acl))) == NULL)
 		failure("calloc");
 
+	rtmp->umask = 0;
 	rtmp->roletype = type;
 	rtmp->rolename = rolename;
 
@@ -337,48 +345,8 @@ add_role_acl(struct role_acl **role, char *rolename, u_int16_t type, int ignore)
 	if (ignore)
 		rtmp->uidgid = special_role_uid++;
 	else if (strcmp(rolename, "default") || !(type & GR_ROLE_DEFAULT)) {
-		if (type & GR_ROLE_USER) {
-			pwd = getpwnam(rolename);
-
-			if (!pwd) {
-				/* now try it as a uid */
-				unsigned long theuid = 0;
-				char *endptr;
-				theuid = strtoul(rolename, &endptr, 10);
-				if (*endptr == '\0')
-					pwd = getpwuid((int)theuid);
-				if (!pwd || theuid > INT_MAX) {
-					fprintf(stderr, "User %s on line %lu of %s "
-						"does not exist.\nThe RBAC system will "
-						"not be allowed to be enabled until "
-						"this error is fixed.\n", rolename,
-						lineno, current_acl_file);
-					exit(EXIT_FAILURE);
-				}
-			}
-
-			rtmp->uidgid = pwd->pw_uid;
-		} else if (type & GR_ROLE_GROUP) {
-			grp = getgrnam(rolename);
-
-			if (!grp) {
-				/* now try it as a gid */
-				unsigned long thegid = 0;
-				char *endptr;
-				thegid = strtoul(rolename, &endptr, 10);
-				if (*endptr == '\0')
-					grp = getgrgid((int)thegid);
-				if (!grp || thegid > INT_MAX) {
-					fprintf(stderr, "Group %s on line %lu of %s "
-						"does not exist.\nThe RBAC system will "
-						"not be allowed to be enabled until "
-						"this error is fixed.\n", rolename,
-						lineno, current_acl_file);
-					exit(EXIT_FAILURE);
-				}
-			}
-
-			rtmp->uidgid = grp->gr_gid;
+		if (type & (GR_ROLE_USER | GR_ROLE_GROUP)) {
+			rtmp->uidgid = get_id_from_role_name(rolename, type);
 		} else if (type & GR_ROLE_SPECIAL) {
 			rtmp->uidgid = special_role_uid++;
 		}
@@ -421,6 +389,35 @@ int count_slashes(char *str)
 static int
 add_globbing_file(struct proc_acl *subject, char *filename,
 		  u_int32_t mode, int type)
+{
+	struct glob_file *glob = malloc(sizeof (struct glob_file));
+	if (!glob)
+		failure("malloc");
+
+	glob->role = current_role;
+	glob->subj = subject;
+	glob->filename = filename;
+	glob->mode = mode;
+	glob->type = type;
+	glob->policy_file = current_acl_file;
+	glob->lineno = lineno;
+	glob->next = NULL;
+
+	
+	if (!glob_files_head) {
+		glob_files_head = glob;
+	} else {
+		glob_files_tail->next = glob;
+	}
+
+	glob_files_tail = glob;
+
+	return 1;
+}
+
+int
+add_globbed_object_acl(struct proc_acl *subject, char *filename,
+		  u_int32_t mode, int type, char *policy_file, unsigned long line)
 {
 	char *basepoint = gr_strdup(filename);
 	char *p, *p2;
@@ -453,8 +450,9 @@ add_globbing_file(struct proc_acl *subject, char *filename,
 
 	if (!anchor) {
 		fprintf(stderr, "Error on line %lu of %s:\n"
-			"Object %s needs to be specified before globbed object %s\n",
-			lineno, current_acl_file, basepoint, filename);
+			"Object %s needs to be specified in the same subject as globbed object %s.\n"
+			"The RBAC system will not be allowed to be enabled until this error is corrected.\n\n",
+			line, policy_file, basepoint, filename);
 		exit(EXIT_FAILURE);
 	}
 
@@ -545,7 +543,9 @@ parse_homedir(char *filename)
 	pwd = getpwuid(current_role->uidgid);
 
 	if (pwd == NULL) {
-		fprintf(stderr, "Error: /etc/passwd was modified during parsing.\n");
+		fprintf(stderr, "Error: Unable to use $HOME on line %lu of %s"
+			", as it can only be used in roles for users that exist"
+			" at the time RBAC policy is enabled.\n", lineno, current_acl_file);
 		exit(EXIT_FAILURE);
 	}
 
@@ -564,6 +564,8 @@ parse_homedir(char *filename)
 	return newfilename;
 }
 
+static uid_t symlink_uid;
+
 int
 add_proc_object_acl(struct proc_acl *subject, char *filename,
 		    u_int32_t mode, int type)
@@ -623,34 +625,19 @@ add_proc_object_acl(struct proc_acl *subject, char *filename,
 
 	file_len++;
 
-	num_objects++;
-	/* one for the object, one for the filename, one for the name entry struct, and one for the inodev_entry struct in the kernel*/
-	num_pointers += 4;
+	memset(&fstat, 0, sizeof(fstat));
 
 	if (lstat64(filename, &fstat)) {
 		dfile = add_deleted_file(filename);
 		fstat.st_ino = dfile->ino;
 		fstat.st_dev = 0;
 		mode |= GR_DELETED;
-		link_count = 0;
-	} else if (S_ISLNK(fstat.st_mode)) {
-		if (link_count > MAX_SYMLINK_DEPTH) {
-			fprintf(stderr, "Error: Too many levels of symbolic links when accessing "
-					"%s\n", filename);
-			exit(EXIT_FAILURE);
-		} else {
-			char buf[PATH_MAX];
-			memset(&buf, 0, sizeof (buf));
-			if (!realpath(filename, buf)) {
-				fprintf(stderr, "Error determining real path for %s\n", filename);
-				exit(EXIT_FAILURE);
-			}
-			link_count++;
-			if(!add_proc_object_acl(subject, gr_strdup(buf), mode, type | GR_SYMLINK))
-				return 0;
-		}
 	}
 
+	num_objects++;
+	/* one for the object, one for the filename, one for the name entry struct, and one for the inodev_entry struct in the kernel*/
+	num_pointers += 4;
+
 	if ((p =
 	     (struct file_acl *) calloc(1, sizeof (struct file_acl))) == NULL)
 		failure("calloc");
@@ -686,7 +673,7 @@ add_proc_object_acl(struct proc_acl *subject, char *filename,
 			return 1;
 		}
 	} else if ((p2 = is_proc_object_dupe(subject, p))) {
-		if (type & GR_SYMLINK)
+		if (p2->mode == p->mode)
 			return 1;
 		fprintf(stderr, "Duplicate object found for \"%s\""
 			" in role %s, subject %s, on line %lu of %s.\n"
@@ -695,14 +682,17 @@ add_proc_object_acl(struct proc_acl *subject, char *filename,
 			subject->filename, lineno, 
 			current_acl_file ? current_acl_file : "<builtin_fulllearn_policy>", p->filename);
 		display_all_dupes(subject, p);
-		fprintf(stderr, "specified on an earlier line.  "
-			"The RBAC system will not load until this"
-			" error is fixed.\n");
+		fprintf(stderr, "specified on an earlier line.\n");
+		fprintf(stderr, "The RBAC system will not load until this error is fixed.\n");
 		exit(EXIT_FAILURE);
 	}
 
 	insert_acl_object(subject, p);
 
+	if (S_ISLNK(fstat.st_mode)) {
+		add_symlink(subject, p);
+	}
+
 	return 1;
 }
 
@@ -745,6 +735,7 @@ add_proc_subject_acl(struct role_acl *role, char *filename, u_int32_t mode, int
 
 	file_len = strlen(filename) + 1;
 
+	// FIXME: for subjects we currently follow symlinks
 	if (stat(filename, &fstat)) {
 		dfile = add_deleted_file(filename);
 		fstat.st_ino = dfile->ino;
@@ -832,6 +823,9 @@ role_mode_conv(const char *mode)
 		case 'A':
 			retmode |= GR_ROLE_GOD;
 			break;
+		case 'R':
+			retmode |= GR_ROLE_PERSIST;
+			break;
 		case 'T':
 			retmode |= GR_ROLE_TPE;
 			break;
@@ -847,13 +841,20 @@ role_mode_conv(const char *mode)
 		}
 	}
 
+	if ((retmode & (GR_ROLE_SPECIAL | GR_ROLE_PERSIST)) == GR_ROLE_PERSIST) {
+		fprintf(stderr, "Error on line %lu of %s.  Persistent "
+			"roles are only valid in the context of special roles.\n"
+			"The RBAC system will not load until this error is fixed.\n", lineno, current_acl_file);
+		exit(EXIT_FAILURE);
+	}
+
 	if ((retmode & (GR_ROLE_NOPW | GR_ROLE_PAM)) == (GR_ROLE_NOPW | GR_ROLE_PAM)) {
 		fprintf(stderr, "Error on line %lu of %s.  The role mode must contain only one of the noauth and pamauth modes.\n", lineno, current_acl_file);
 		exit(EXIT_FAILURE);
 	}
 
-	if (retmode & GR_ROLE_SPECIAL &&
-	    retmode & (GR_ROLE_USER | GR_ROLE_GROUP)) {
+	if ((retmode & GR_ROLE_SPECIAL) &&
+	    (retmode & (GR_ROLE_USER | GR_ROLE_GROUP))) {
 		fprintf(stderr, "Error on line %lu of %s.  The role mode must be either "
 				"special, or user/group, not both.\n"
 				"The RBAC system will not load until this"
@@ -941,6 +942,12 @@ proc_subject_mode_conv(const char *mode)
 		case 'a':
 			retmode |= GR_KERNELAUTH;
 			break;
+		case 's':
+			retmode |= GR_ATSECURE;
+			break;
+		case 'x':
+			retmode |= GR_SHMEXEC;
+			break;
 		default:
 			fprintf(stderr, "Invalid subject mode "
 				"\'%c\' found on line %lu "
@@ -1027,6 +1034,18 @@ proc_object_mode_conv(const char *mode)
 		case 's':
 			retmode |= GR_SUPPRESS;
 			break;
+		case 'f':
+			if (!(current_role->roletype & GR_ROLE_PERSIST)) {
+				fprintf(stderr, "Error on line %lu of "
+				"%s.  The 'f' mode is only permitted "
+				"within persistent special roles.\n"
+				"The RBAC system will not be allowed to "
+				"be enabled until this error is corrected.\n",
+				lineno, current_acl_file);
+				exit(EXIT_FAILURE);
+			}
+			retmode |= GR_INIT_TRANSFER;
+			break;
 		case 'm':
 			retmode |= GR_SETID;
 			break;
